/** 
 *  Yudit Unicode Editor Source File
 *
 *  Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "stoolkit/STextData.h"

#include "stoolkit/SUniMap.h"
#include "stoolkit/SEncoder.h"

/**
 * @param attr is tru if no text was added or removed. only attribute changed.
 */
STextDataEvent::STextDataEvent (const STextIndex& _start, bool attr)
{
  start = _start;
  valid = false;
  attribute = attr;
}
/**
 * Create a stoolkit STextDataEvent
 */
STextDataEvent::STextDataEvent (void)
{
  clear ();
}

/**
 * Set the remaining.
 */
void
STextDataEvent::setRemaining (const STextIndex& _remaining)
{
  remaining = _remaining;
  valid = true;
}

/**
 * Nothing to destroy. Just in case
 */
STextDataEvent::~STextDataEvent ()
{
}

/**
 * Clear all fields.
 */
void
STextDataEvent::clear ()
{
  start = STextIndex (0,0);
  remaining =  STextIndex (0,0);
  valid = false;
  attribute = true;
}

/**
 * Add a new event to this in a sane way.
 */
void
STextDataEvent::add (const STextDataEvent& event)
{
  if (!event.valid)
  {
    //fprintf (stderr, "NONO in STextDataEvent::add\n");
    return;
  }
  if (!valid)
  {
    start = event.start;
    remaining = event.remaining;
    if (!event.attribute) attribute = false;
  }
  else if (event.remaining < remaining)
  {
    remaining = event.remaining;
  }
  valid = true;
}

/**
 * This is the definition of one glyph. It may be composed of
 * several characters.
 * @param _chars is the character array in order.
 * @param _lr is the direction.
 */
SGlyph::SGlyph (const SV_UCS4& _chars, bool _lr)
{
  selected = false;
  underlined = false;

  useComposition = false;
  composition = 0;

  lr = _lr;

  lineend = _chars[0] == LS || _chars[0] == PS 
       || _chars[0] == CR || _chars[0] == LF;
  tab = _chars.size()==1 && _chars[0] == TAB;

  ucs4v.append (_chars);
  if (ucs4v.size()==1) composition = ucs4v[0];
}


const SS_UCS4*
SGlyph::decomposition() const
{
  return ucs4v.array();
}

/**
 * check if this character is special - lineend or tab
 * that requires special rendering
 */
bool
SGlyph::isSpecial() const
{
  return (lineend || tab || ucs4v[0] < 0x20 || 
      (composition != 0 && composition < 0x20));
}

/**
 * Construct a key from characters inside the glyph
 * @return the key. the key does not contain attributes, 
 * it only has glyph info.
 */
SString
SGlyph::charKey() const
{
  return SString ((char*) ucs4v.array(), ucs4v.size() * sizeof (SS_UCS4));
}

bool
SGlyph::operator == (const SGlyph& g2) const
{
  if (g2.size() != ucs4v.size()) return false;
  if (composition != 0 && composition == g2.composition) return true;
  for (unsigned int i=0; i<ucs4v.size(); i++)
  {
     if (ucs4v[i] != g2.ucs4v[i]) return false;
  }
  return true;
}

bool
SGlyph::operator != (const SGlyph& g2) const
{
  if (g2.size() != ucs4v.size()) return true;
  if (composition != 0 && composition != g2.composition) return true;
  for (unsigned int i=0; i<ucs4v.size(); i++)
  {
     if (ucs4v[i] != g2.ucs4v[i]) return true;
  }
  return false;
}

bool
SGlyph::operator == (const SType& _type) const
{
  if (ucs4v.size() != 1) return false;
  return (SS_UCS4) _type == ucs4v[0];
}

bool
SGlyph::operator != (const SType& _type) const
{
  if (ucs4v.size() != 1) return true;
  return (SS_UCS4) _type != ucs4v[0];
}

/**
 * Create a glyph from another glyph
 * @param glyph is the other glyph.
 */
SGlyph::SGlyph (const SGlyph& glyph)
{
  lr = glyph.lr;
  underlined = glyph.underlined;
  selected = glyph.selected;
  lineend = glyph.lineend;
  tab = glyph.tab;
  composition = glyph.composition;
  useComposition = glyph.useComposition;
  ucs4v = glyph.ucs4v;
}

/**
 * Nothing to desctruct now.
 */
SGlyph::~SGlyph ()
{
}

/**
 * Assign a glyph.
 * @param glyph is the other glyph.
 */
SGlyph
SGlyph::operator=(const SGlyph& glyph)
{
  lr = glyph.lr;
  underlined = glyph.underlined;
  selected = glyph.selected;
  lineend = glyph.lineend;
  tab = glyph.tab;
  composition = glyph.composition;
  useComposition = glyph.useComposition;
  ucs4v = glyph.ucs4v;
  return *this;
}

/**
 * All objects are to define this.
 */
SObject*
SGlyph::clone() const
{
  return new SGlyph (*this);
}

unsigned int
SGlyph::size() const
{
  return ucs4v.size();
}

SS_UCS4
SGlyph::operator[] (unsigned int index) const
{
  return ucs4v[index];
}

SV_UCS4
SGlyph::decompose() const
{
  SV_UCS4 ret;
  if (useComposition && composition != 0)
  {
    ret.append (composition);
    return SV_UCS4(ret);
  }
  for (unsigned int i=0; i<ucs4v.size(); i++)
  {
    ret.append (ucs4v[i]);
  }
  
  return SV_UCS4(ret);
}

bool
SGlyph::isWhiteSpace() const
{
  if (isSpecial ()) return true;
  if (ucs4v.size() > 1) return false;

  switch (ucs4v[0])
  {
  case 0x20:
  case 0x3000:
    return true;
  }
  return false;
}

/*------------------------- STextData ---------------------------*/
/**
 * Create and empty STextData 
 */
STextData::STextData (void)
{
}

/**
 * Create a text data from utf8.
 * @param utf8 is the input text
 */
STextData::STextData (const SString& utf8)
{
  insert (utf8);
}

/**
 * clear the data and assign new data
 */
STextData
STextData::operator = (const STextData & data)
{
  clear ();
  insert (data);
  return *this;
}

/**
 * Glyph Lines are pointers
 */
STextData::~STextData ()
{
  for (unsigned int i=0; i<lines.size(); i++)
  {
     delete lines[i];
  }
}

/**
 * Get the contained text
 */
SString
STextData::getText () const
{
  if (size()==0) return SString();
  return getText (STextIndex (0, 0), STextIndex (size(), size(size()))); 
}

/**
 * Get the text between the cursor and here
 */
SString
STextData::getText (const STextIndex& index) const
{
  return getText (textIndex, index); 
}

/**
 * Extract a bi-di marked utf-8 text.
 * @param begin is the starting index
 * @param end is the ending index
 */
SString
STextData::getText (const STextIndex& begin, const STextIndex& end) const
{
  SString ret;
  if (size() == 0) return SString(ret);

  STextIndex start;
  STextIndex stop;
  if (end > begin)
  {
    start = begin;
    stop = end;
  }
  else
  {
    start = end;
    stop = begin;
  }
  if (stop.line > size())
  {
     stop.line = size(); stop.index = size(size()-1);
  }
  for (unsigned int i=start.line; i<=stop.line && i<size(); i++)
  {
     unsigned int b = (i==start.line) ? start.index : 0;
     unsigned int e = (i==stop.line)? stop.index : size (i);
     ret.append (getText (i, b, e));
  }
  return SString(ret);
}

/**
 * Extract a bi-di marked utf-8 text.
 * TODO: implements a better bidi.
 * @param beginLine is the starting line
 * @param beginIndex is the starting position
 * @param endIndex is the ending position
 */
SString
STextData::getText (unsigned int beginLine, unsigned int beginIndex,
   unsigned int endIndex) const
{
  SV_UCS4 ucs;

  for (unsigned int i=beginIndex; i<endIndex; i++)
  {
     if (glyphAt(STextIndex (beginLine, i)).lr)
     {
          ucs.append (glyphAt(STextIndex (beginLine, i)).decompose());
     }
     else
     {
        unsigned int start = i;
        while (i<endIndex && !glyphAt(STextIndex (beginLine, i)).lr)
        {
           i++;
        }
        unsigned int end = i;
        i--; /* We will increment it later in the external loop */
        ucs.append (SGlyph::RLO);
        while (end>start)
        {
          --end;
          ucs.append (glyphAt(STextIndex (beginLine, end)).decompose());
        }
        ucs.append (SGlyph::PDF);
     }
  }
  SEncoder ic;
  return (SString(ic.encode (ucs)));
}

/**
 * find string in text data from current position.
 * return an index to the end of that data.
 * move the current position to the beginning of the data.
 * it does not search across lines.
 */
STextIndex
STextData::find (const SString& string)
{
  STextData d (string);

  /**
   * go through all lines.
   */
  for (unsigned int i=textIndex.line; i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = size (i);

     SGlyphLine* line = lines[i];
     /**
      * go through all glyphs
      */
     for (unsigned int j=begin; j+d.size(0)<=end; j++)
     {
        /* hack. We dont disclose the SGlyphLIne= */
        bool found = true;
        for (unsigned int k=0; k<d.size(0); k++)
        {
          const SGlyph* g = line->peek(j+k);
          const SGlyph& g2 = d.glyphAt (STextIndex(0, k));
          if (g2 != *g)
          {
            found = false;
          }
        }
        if (found)
        {
          move (STextIndex(i, j));
          return STextIndex (i, j+d.size(0));
        }
     }
  }
  return STextIndex (0,0);
}

/**
 * Get a glyph at a certain position
 * @param index is the glyph index
 * @return a reference to the glyph.
 */
const SGlyph&
STextData::glyphAt (const STextIndex& index) const
{
  return (*lines[index.line])[index.index];
}

/**
 * move the insertion point.
 * @param line is the line number.
 * @param index is the line index.
 */
void
STextData::move (const STextIndex& index)
{
  STextIndex ndx (index);
  if (ndx.line > size()) ndx.line = size();
  if (ndx.line == size())
  {
    ndx.index = 0;
  }
  else if (ndx.index >= size(ndx.line))
  {
    ndx.index = size(ndx.line);
    if (isProperLine (ndx.line) && ndx.index == size(ndx.line))
    {
      ndx.index = ndx.index-1;
    }
  }
  textIndex = ndx;
}

/**
 * set the text data
 */
void
STextData::setText (const SString& string)
{
  clear ();
  insert (string);
}

/**
 * insert one signle glyph and move the cursor to the right
 * @param glyph is the glyph to insert
 * @return an index to end, text between current cursor and end
 * has been inserted.
 */
void
STextData::insert (const SGlyph& glyph, STextIndex* ret)
{
  STextIndex oldIndex = textIndex;

  if (size() <= textIndex.line)
  {
     lines.append (new SGlyphLine());
  }

  SGlyphLine* line = lines[textIndex.line];
  line->insert (textIndex.index, glyph);
  if (ret->line == textIndex.line)
  {
    ret->index++;
  }
  textIndex.index = textIndex.index+1;

  STextDataEvent nevt (oldIndex);
  /* there are two cases. we opened a new line or not */
  if (glyph.lineend)
  {
    /* cut line in two */
    if (line->size() > textIndex.index)
    {
      SGlyphLine *newLine = new SGlyphLine (*line);
      newLine->remove (0, textIndex.index);
      line->truncate (textIndex.index);
      lines.insert (textIndex.line+1, newLine);

      textIndex.index = newLine->size()-1;
      textIndex.line = textIndex.line + 1;
      setMaxLimits (&nevt, textIndex);
      textIndex.index = 0;
      ret->line++;
      if (ret->line== textIndex.line) ret->index=0;
    }
    else /* new line */
    {
      textIndex.index = 0;
      textIndex.line = textIndex.line + 1;
      setMaxLimits (&nevt, textIndex);
      ret->line++;
      ret->index=0;
      if (ret->line== textIndex.line) ret->index=0;
    }
  }
  else
  {
    setMaxLimits (&nevt, textIndex);
  }
  //fprintf (stderr, "Adding: %u %u - %u %u textevent.\n", 
   //   nevt.start.line, nevt.start.index, nevt.remaining.line, nevt.remaining.index);
  event.add (nevt);
  //fprintf (stderr, "Adding: %u %u - %u %u became.\n", 
   //   event.start.line, event.start.index, event.remaining.line, event.remaining.index);
}

/**
 * Insert a text from data
 * @param data is the STextData
 */
void
STextData::insert (const STextData& data)
{
  STextIndex ind;
  for (unsigned int i=0; i<data.size(); i++)
  {
    for (unsigned int j=0; j<data.size(i); j++)
    {
      insert (data.glyphAt (STextIndex (i, j)), &ind);
    }
  }
}

/**
 * Insert a text from utf8. Do fuzzy algoritm to determine
 * direction and all kinds of fuzzy properties, like composition.
 * The resulting text is added to the textbuffer in screen-order.
 * FIXME: 
 * - Implement a bidi algoritm. The current version treats everything
 *   as strong left-to-right character unless it is enclosed in  RLO-PDF.
 * - Get a table of all composing characters and break down the glyphs so
 *   that one glyph contains the base character and the compositng marks.
 * @param utf8 an utf8 encoded text.
 * @return an index to end, text between current cursor and end
 * has been inserted. 
 */
STextIndex
STextData::insert (const SString& utf8)
{
  SEncoder ic;
  SV_UCS4 ucs4 = ic.decode (utf8, false);
  /* split into lines */
  unsigned int start = 0;
  unsigned int end = 0;
  STextIndex old=textIndex;
  do
  {
    SGlyphLine gl;
    start = end;

    /* split the lines and do stoolkit composition */
    end = split (ucs4, &gl, start);

    /* put bidi into OUR visual order */
    bidiToVisual (&gl);

    for (unsigned int i=0; i<gl.size(); i++)
    {
      insert (gl[i], &old);
    }
  } while (start != end);

  return STextIndex (old);
}

/**
 * remove glyphs between current and line and size.
 * current can be less or greater.
 * @param index is the ending high end is non-exclusive
 */
void
STextData::remove (const STextIndex& index)
{
  if (size() == 0) return;
  //fprintf (stderr, "remove: %u %u %u %u\n", 
  //    index.line, index.index, textIndex.line, textIndex.index);

  STextIndex newIndex = reorder (index);
  //fprintf (stderr, "remove reorder[0]: %u %u\n", newIndex.line, newIndex.index);
  //fprintf (stderr, "remove reorder[1]: %u %u\n", textIndex.line, textIndex.index);
  unsigned int i;
  bool join = false;
  for (i=textIndex.line; i<=newIndex.line && i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = (i==newIndex.line)? newIndex.index : size (i);

     SGlyphLine* line = lines[i];
    //fprintf (stderr, "REMOVE: %u %u - %u \n", i, begin, end);
     if (begin == 0 && end == size(i))
     {
        if (isProperLine(i)) join = true;
        line->clear();
        //fprintf (stderr, "replacing full ine %u\n", i);
     }
     else for (unsigned int j=begin; j<end; j++)
     {
        /* hack. We dont disclose the SGlyphLIne= */
        if (j+1==end)
        {
          SGlyph g = glyphAt (STextIndex(i, begin));
          if (g.lineend) join = true;
        }
        line->remove (begin);
     }
  }
  unsigned int lr = 0;

  /* Remove zeroes */
  for (i=textIndex.line+1; i<=newIndex.line && i<size() + lr; i++)
  {
     SGlyphLine* line = lines[i-lr];
     if (line->size() == 0)
     {
       //fprintf (stderr, "removing full ine %u\n", i);
       lines.remove (i-lr);
       delete line;
       lr++;
     }
  }

  /* now we have everything collapsed. */

  STextDataEvent nevt (textIndex);
  setMaxLimits (&nevt, textIndex);
  event.add (nevt);

  /* a joining line was created */
  if (join)
  {
    /* wanted, but impossible */
    if (textIndex.line+1 == size())
    {
      SGlyphLine* line = lines[textIndex.line];
      if (line->size()==0)
      {
        lines.remove (textIndex.line);
        delete line;
      }
      return;
    }
    //fprintf (stderr, "join %u\n", textIndex.line+1);
    SGlyphLine* gl = lines[textIndex.line+1];
    lines.remove (textIndex.line+1);

    STextIndex sv = textIndex;
    STextIndex dm = textIndex;
    for (i=0; i<gl->size(); i++)
    {
      insert ((*gl)[i], &dm);
    }
    delete gl;
    move (sv);
  }
  else
  { /* we may not removed this */
    SGlyphLine* line = lines[textIndex.line];
    if (line->size()==0)
    {
      lines.remove (textIndex.line);
      delete line;
    }
  }
}

/**
 * select a region of text and move cursor to it.
 * @param index is the new index. high end is non-exclusive
 * @param is is true if select, unselect otherwise. 
 */
void
STextData::select (const STextIndex& index, bool is)
{
  if (size() == 0) return;

  STextIndex newIndex = reorder (index);
  bool reordered = (newIndex != index);


  if (newIndex.line > size())
  {
     newIndex.line = size(); newIndex.index = size(size()-1);
  }
  for (unsigned int i=textIndex.line; i<=newIndex.line && i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = (i==newIndex.line)? newIndex.index : size (i);

     SGlyphLine* line = lines[i];
     for (unsigned int j=begin; j<end; j++)
     {
        /* hack. We dont disclose the SGlyphLIne= */
        SGlyph* g = (SGlyph*) line->peek(j);
        g->selected = is;
     }
  }
//fprintf (stderr, "reordered = %d. TextIndex=%d, NewIndex=%d\n", 
//  (int) reordered, textIndex.line, newIndex.line);
  if (reordered)
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
  }
  else
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
    textIndex =  newIndex;
  }
}

/**
 * select a region of text and move cursor to it.
 * @param index is the new index - high end is non-exclusive
 * @param is is true if underline, un-underline otherwise.
 */
void
STextData::underline (const STextIndex& index, bool is)
{
  if (size() == 0) return;

  STextIndex newIndex = reorder (index);
  bool reordered = (newIndex != index);

  for (unsigned int i=textIndex.line; i<=newIndex.line && i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = (i==newIndex.line)? newIndex.index : size (i);

     SGlyphLine* line = lines[i];
     for (unsigned int j=begin; j<end; j++)
     {
        /* hack. We dont disclose the SGlyphLIne= */
        SGlyph* g = (SGlyph*) line->peek(j);
        g->underlined = is;
     }
  }
  if (reordered)
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
  }
  else
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
    textIndex =  newIndex;
  }
}

/**
 * Clear the text
 */
void
STextData::clear ()
{
  for (unsigned int i=0; i<lines.size(); i++)
  {
     delete lines[i];
  }
  lines.clear();
  textIndex = STextIndex (0, 0);
  STextDataEvent nevt (STextIndex (0, 0));
  nevt.remaining = STextIndex (0, 0);
  nevt.valid = true;
  event.add (nevt);
}

void
STextData::fireEvent ()
{
  if (!event.valid || listener == 0) return ;
  listener->textChanged (this, event); 
  event.clear();
}

void
STextData::clearEvent ()
{
  event.clear();
}

/**
 * return the number of lines in text
 */
unsigned int
STextData::size() const
{
  return lines.size();
}

/**
 * return the number of lines in line
 */
unsigned int
STextData::size(unsigned int line) const
{
  if (line >= size()) return 0;
  const SGlyphLine* l = lines[line];
  return l->size();
}

/**
 * Get the text index relative to offset.
 * @param offset is character offset to textIndex.
 */
STextIndex
STextData::getTextIndex(int charOffset) const
{
  STextIndex ret =  textIndex;
  if (charOffset > 0)
  {
    for (int i=0; i<charOffset; i++)
    {
      if (ret == STextIndex (size(), 0)) break;
      if (size(ret.line) <= ret.index+1)
      {
        if (!isProperLine (ret.line))
        {
          if (size(ret.line) == ret.index+1)
          {
             ret.index = ret.index + 1;
          }
          break;
        }
        ret.index = 0;
        ret.line = ret.line + 1;
        continue;
      }
      ret.index = ret.index + 1;
      continue;
    }
  }
  else if (charOffset < 0)
  {
    for (int i=0; i>charOffset; i--)
    {
      if (ret == STextIndex (0, 0)) break;
      if (ret.index > 0)
      {
        ret.index = ret.index - 1;
        continue;
      }
      ret.line = ret.line-1;
      /*
      * For proper lines,
      * The beginning of the line is the same location as the end. 
      */
      ret.index = size(ret.line);
      if (isProperLine (ret.line))
      {
        ret.index = ret.index-1;
      }
    }
     
  }
  return STextIndex (ret);
}

/**
 * Check if line ends with newline glyph
 */
bool
STextData::isProperLine (unsigned int line) const
{
  if (line >= size()) return false;
  if (size(line)==0) return false;
  const SGlyph g = glyphAt(STextIndex(line, size(line)-1));
  return g.lineend;
}

/**
 * add data listener.
 * TODO: now it only sets it.
 */
void
STextData::addTextDataListener (STextDataListener* _listener)
{
  listener = _listener;
}

/**
 * Set the maximum limits.
 * @param evt -> remainingLine and remainingPosition will be set
 * @param index is the one that needs to be converted to remaining index
 */
void
STextData::setMaxLimits (STextDataEvent* evt, const STextIndex& index)
{
  evt->valid = true;
  unsigned int lsize = size();
  if (index.line >= lsize)
  {
     evt->setRemaining (STextIndex (0,0));
     return;
  }
  unsigned int rem  = lsize - index.line - 1;
  unsigned int cindex = size(index.line);
  evt->setRemaining (STextIndex (rem, (index.index>=cindex) ? 0 : cindex - index.index));
}

/**
 * Convert remainin-line to maxline
 */
STextIndex
STextData::getMaxTextIndex (const STextDataEvent& evt) const
{
  if (!evt.valid) return STextIndex (0,0);
  unsigned int lsize = size();
  if (lsize <= evt.remaining.line)
  {
     return STextIndex (lsize,0);
  }
  unsigned int line = lsize - evt.remaining.line -1;
  if (size(line) < evt.remaining.index)
  {
     return STextIndex (line, size(line));
  }
  return STextIndex (line, size(line)+1-evt.remaining.index);
}

/**
 * Put the stuff in order, so that it will always be in increasing order.
 * @param end is the desired end
 */
STextIndex
STextData::reorder (const STextIndex& index)
{
  if (index > textIndex)
  {
     return STextIndex (index);
  }
  STextIndex tmp = textIndex;
  textIndex = index;
  return STextIndex (tmp);
}  

/**
 * return true if character is a target for select.
 */
bool
STextData::isWhiteSpace (const STextIndex& index) const
{
  return glyphAt (index).isWhiteSpace();
}

/**
 * Split a text into lines
 * @param ucs4 is a text that can contain newline chars.
 * @param from is the starting index.
 * @parem gl is the return glyphline.
 * @return ending index, that is equal to starting index 
 * if ther is no more data.
 */
unsigned int
split (const SV_UCS4& ucs4, SGlyphLine* gl, unsigned int from)
{
  SUniMap composer ("precompose");
  SUniMap category ("category");
  gl->clear();
  unsigned int i=from;
  SS_UCS4 composition;
  bool    useComposed;
  while (i<ucs4.size())
  {
    composition = 0;
    useComposed = false;
    SV_UCS4 du4;
    switch (ucs4[i])
    {
    case SGlyph::CR:
      du4.append (ucs4[i++]);
      if (i<ucs4.size() && ucs4[i] == SGlyph::LF)
      {
        du4.append (ucs4[i++]);
      }
      break;
    case SGlyph::LF:
      du4.append (ucs4[i++]);
      if (i<ucs4.size() && ucs4[i] == SGlyph::CR)
      {
        du4.append (ucs4[i++]);
      }
      break;
    case SGlyph::PS:
    case SGlyph::LS:
    case SGlyph::RLO:
    case SGlyph::LRO:
    case SGlyph::PDF:
      du4.append (ucs4[i++]);
      break ;
    default:
      /* a quick hack to test composing Hebrew for Raphael */
      if (composer.isOK () && category.isOK())
      {
        SV_UCS4 ret;
        unsigned int n = composer.lift (ucs4, i, false, &ret);
        if (ret.size()==1) composition = ret[0];

#if 0
if (ucs4[i] == 0x05D0) // with 05B7
{
  fprintf (stderr, "0x05D0=");
}
else if (ucs4[i] == 0xFB2E)
{
  fprintf (stderr, "0xFB2E=");
}
else if (ucs4[i] == 0x202D)
{
  fprintf (stderr, "0x202D=");
}
#endif

        if (n==i) /* it was not a composition */
        {
          ret.clear();
         /* check if it can be broken down into a composition */
          unsigned int n = composer.lift (ucs4, i, true, &ret);
          if (ucs4[i] != 0 && n == i+1 
              && ret.size() > 0 && ret.size() <= SD_MAX_COMPOSE)
          {
//fprintf (stderr, "use composed...\n");
            useComposed = true;
            n = i+1;
            
            composition = ucs4[i++];
            du4.append (ret);
          }
          else
          {
            du4.append (ucs4[i++]);
            /* check if next is composing, support this level only */
            while (i<ucs4.size())
            {
               SS_UCS2 u2 = category.encode (ucs4[i]);
#if 0
if (ucs4[i] == 0x0e34)
{
fprintf (stderr, "0x0e34=%u\n", (unsigned int) u2);
}
#endif
               if (u2 == 0x4) // Mark, Non-Spacing
               {
//fprintf (stderr, "decomposed 0x%x\n", (unsigned int) ucs4[i]);
                  du4.append (ucs4[i++]);
               }
               else
               {
                  break;
               }
            }
          }
        }
        else
        {
          while (i<n)
          {
            du4.append (ucs4[i++]);
          }
        }
      }
      else
      {
        du4.append (ucs4[i++]);
      }
      break;
    }
    SGlyph g(du4);
    g.useComposition = useComposed;
    g.composition = composition;
    gl->append (g);
    if (g.lineend)
    {
      /* line can not start with pdf. */
      if (i<ucs4.size() && ucs4[i]==SGlyph::PDF) i++;
      break;
    }
  }
  return i;
}

/**
 * put the text into visual order.
 * @param gl is a line of glyphs. 
 * TODO: make a bettter bidi.
 */
void
bidiToVisual (SGlyphLine* gl)
{
  SGlyphLine in = *gl;
  gl->clear();
  SBinVector<bool> stack;
  stack.append (true);
  bool lrnow = true;
  unsigned int rlends=0;

  /*
   * go through the input and construct gl.
   */
  for (unsigned int i=0; i<in.size(); i++)
  {
    //fprintf (stderr, "in[%u] = %x\n", i, (unsigned int) (in[i][0]));
    if (in[i] == SGlyph::LRO)
    {
       stack.append (true);
       lrnow = true;
    }
    else if (in[i] == SGlyph::RLO)
    {
       stack.append (false);
       lrnow = false;
       for (rlends=i+1; rlends<in.size(); rlends++)
       {
          if (in[rlends] == SGlyph::LRO || in[rlends] == SGlyph::RLO
              || in[rlends] == SGlyph::PDF)
          {
             break;
          }
       }
    }
    else if (in[i] == SGlyph::PDF)
    {
       if (stack.size() > 1)
       {
         lrnow = stack[stack.size()-2];
         stack.truncate (stack.size()-1);
       }
       else
       {
         fprintf (stderr, "STextLine::insert unmatched STextType::PDF\n");
         lrnow = true;
       }
       if (!lrnow) 
       {
         for (rlends=i+1; rlends<in.size(); rlends++)
         {
           if (in[rlends] == SGlyph::LRO || in[rlends] == SGlyph::RLO
                || in[rlends] == SGlyph::PDF)
           {
              break;
           }
         }
       }
    }
    else
    {
       if (lrnow)
       {
         SGlyph g (in[i]);  
         g.lr = lrnow;
         gl->append (g);
       }
       else
       {
         rlends--;
         SGlyph g (in[rlends]);
         g.lr = lrnow;
         gl->append (g);
       }
    }
  }
}

#if 0
/* FIXME: move it to SString */
static void
truncateUTF8 (SString * str, unsigned int size)
{
  unsigned int k = 0;
  unsigned int i = 0;
  while (i<str->size() && k < size)
  {
     unsigned int utfsize = 1;
     if (((unsigned char) (*str)[i] & 0xe0) ==0xc0)
     {
        utfsize = 2;
     }
     else if (((unsigned char) (*str)[i] & 0xf0) ==0xe0)
     {
        utfsize = 3;
     }
     i += utfsize;
     k++;
  }
  if (i>=str->size())
  {
    return;
  }
  else if (i==0)
  {
     str->clear();
     return;
  }
  str->truncate (i);
  return;
}      
#endif
