/** 
 *  Yudit Unicode Editor Source File
 *
 *  Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef USE_GETTEXT
extern "C"
{
#include <libintl.h>
}
#endif
#include <stdlib.h>

#include "stoolkit/SUtil.h"
#include "stoolkit/STypes.h"
#include "stoolkit/SString.h"
#include "stoolkit/SStringVector.h"
#include "stoolkit/SProperties.h"
#include "stoolkit/SIO.h"
#include "stoolkit/SIOStream.h"
#include "stoolkit/SEncoder.h"
#include "swindow/SAwt.h"

#ifndef USE_WINAPI
#include <unistd.h> 
#include <pwd.h>
#include <sys/types.h>
#endif



SString getHome ()
{
#ifndef USE_WINAPI
  struct passwd * p = getpwuid(getpid());
  if (p!=0)
  {
    return SString (p->pw_dir);
  }
#endif
  char* henv = getenv ("HOME");
  if (henv)
  {
    return SString(henv);
  }
#ifdef USE_WINAPI
  fprintf (stderr, "can not determine home directory. using C:\\temp");
  return SString ("C:\\temp");
#else
  fprintf (stderr, "can not determine home directory. /tmp");
  return "/tmp";
#endif
}

#ifndef YUDIT_DATA
#define YUDIT_DATA "/usr/share/yudit"
#endif YUDIT_DATA 

SString getPrefix ()
{
   return YUDIT_DATA;
}

bool
loadProperties (const SString& file, SProperties* in)
{
  SFile f(file);
  if (f.size() < 0) return false;
  SFileImage im = f.getFileImage ();
  if (im.size()<0) return false;
  SProperties p(SString (im.array(), im.size()));
  in->merge(p);
  return true;
}

bool
loadProperties (SProperties* in)
{
  SString c1 = getHome();
  c1.append ("/.yudit/yudit.properties");
  SString c2 = getPrefix();
  c2.append ("/config/yudit.properties");

  bool ret = loadProperties (c2, in);
  ret = ret || loadProperties (c1, in);
  return ret;
}

bool
saveProperties (const SString& file, const SProperties& out)
{
  SProperties orig;
  loadProperties (file, &orig);
  orig.merge (out);
  SFile f(file);
  SOutputStream os = f.getOutputStream();
  if (!os.isOK()) return false;

  SWriter writer(os);
  if (!writer.isOK()) return false;
  SString sh("#Yudit-");
  sh.append (SD_YUDIT_VERSION);
  sh.append (" properties\n#\n");
  if (!writer.write (sh)) return false;
  SString s(orig.toString());
  s.append ("\n#end of Yudit-");
  s.append (SD_YUDIT_VERSION);
  s.append (" properties\n");
  if (!writer.write (s)) return false;
  if (!writer.close()) return false;
  return true;
}

static bool translateInited = false;

void
initTranslate (const SString& bindIn, const SString& domain)
{
  translateInited=true;
#ifdef USE_GETTEXT
  SString in(bindIn); in.append ((char)0);
  SString dom(domain); dom.append ((char)0);
  bindtextdomain (in.array(), dom.array());
  textdomain (in.array());
  return;
#else
  return;
#endif
}

void
setLanguage (const SString& _str)
{
  static char*  lang0=0;
  static char*  lang1=0;

  if (!translateInited) return;
  SString str(_str);
  str.append ((char)0);

#if 0
  setlocale (LC_MESSAGES, lang_);
#endif

  if (lang1!=0 && !strcmp (lang1+strlen("LANGUAGE="), str.array())) return;

  // putenv just eats it...
  //if (lang0) delete lang0;
  //if (lang1) delete lang1;

  lang0 = new char[strlen ("LANG=") + strlen (str.array()) + 1];
  strcpy (lang0, "LANG=");
  strcat (lang0, str.array());
  putenv (lang0);

  lang1 = new char[strlen ("LANGUAGE=") + strlen (str.array()) + 1];
  strcpy (lang1, "LANGUAGE=");
  strcat (lang1, str.array());
  putenv (lang1);
  //textdomain ("messages");
  return;
}

SString getLanguage ()
{
  const char* l = getenv ("LANG");
  if (l==0) l = getenv ("LANGUAGE");
  if (l==0) return SString();
  return SString (l);
  
}

SString translate (const SString& str)
{
  if (str.size()==0) return SString(str);
#ifdef USE_GETTEXT
  SString nstr(str);
  nstr.append ((char)0);
  if (translateInited) return SString(gettext (nstr.array()));
  return SString(str);
#else
  return SString(str);
#endif
}

/**
 * print the unicode value in "U+XXXX or U+XXXXXXXX" form
 */
SString
unicodeValueOf (const SV_UCS4& ucs4)
{
  char a[64];
  SString s;
  for (unsigned int i=0; i<ucs4.size(); i++)
  {
    if (i!=0) s.append (" ");
    if (ucs4[i] > 0xffff)
    {
      sprintf (a, "%08lX", (unsigned long) ucs4[i]);
    }
    else
    {
      sprintf (a, "%04X", (unsigned int) ucs4[i]);
    }
    s.append (a);
  }
  return SString (s);

}

SString
unicodeValueOf (const SString& str)
{
  SEncoder enc("utf-8");
  SV_UCS4 ucs4 = enc.decode (str);
  return unicodeValueOf (ucs4);
}
