/** 
 *  Yudit Unicode Editor Source File
 *
 *  Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 

// This file defines several UCS2Converter classes
//

#include "UString.h"
#include "UCS2Conv.h"
#include "UCommon.h"
#include <memory.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include "NMap.h"
#include "UMap.h"

static unsigned char *allowedChars = 0;
static unsigned char *base64Decode = 0;

//
// 8 Bit converters with umap
//
U8BitConv::U8BitConv (const char* mapName) : UCS2Converter (mapName)
{
	initialized = 0;
	map=0;
}

U8BitConv::~U8BitConv()
{
	if (map) delete map;
}

void
U8BitConv::initialize ()
{
	if (initialized) return;

	map = new UMap ();
	CHECKNULL (map);
	if (map->rename(name) != UMap::OK)
	{
		cerr << "error: can not get " << name << " umap.\n";
		delete map;
		map =0;
	}
	initialized = 1;
}

// 
// Only chars that are >= 0x80 are mapped
// IN case the encoder can map > 0xff two byte stream is assumed.
// In case it is 16bit mapped we use Big endian.
//
UCS2*
U8BitConv::decode (const unsigned char* in, int _size, int* retSize)
{
	UCS2		got;
	int		i;
	int		step;
	UCS2		value;

	initialize (); 
	init (in, _size);

	if (map==0)
	{
		for (i=0; i<inputSize; i++)
		{
			value = (UCS2) in[i];
			append (value);
		}
		terminateUCS2 ();
		*retSize = size-1;
		return ucs2;
	}

	step = (map->getDecodeInWordSize() > 0) ? 2 : 1;

	NStateModel model(1, map->getDecodeInWordSize (),
		map->getDecodeOutWordSize (), 1);

	NBuffer	buffer(1,1);
	int stt, j;

	for (i=0; i<inputSize-step+1; i=i+step)
	{
		value = (step==1) ? in[i] : ((in[i]<<8) | in[i+1]); 
		buffer.append (&value, 1);
		stt = map->decode (&value, 1, &model, (i<inputSize+1-2*step) ? 1 : 0);
		if (stt == NM_ACCEPT)
		{
			continue;
		}
		if (stt == NM_REJECT)
		{
			model.out.reset ();
			value=(UCS2) buffer.u.u16[0];
			append (value);
			i = i - (buffer.length-1);
			buffer.reset ();
			continue;
		}
		for (j=0; j<model.out.length; j++)
		{
			value=(UCS2) model.out.u.u16[j];
			append (value);
		}
		i = i - (buffer.length-stt);
		model.reset();
		buffer.reset();
	}
	terminateUCS2 ();
	*retSize = size-1;
	return ucs2;
}

// 
// Only chars that are >= 0x80 are mapped
// IN case the encoder can map > 0xff two byte stream is assumed.
// In case it is 16bit mapped we use Big endian.
//
unsigned char*
U8BitConv::encode (const UCS2 *in, int _size, int* retSize)
{
	UCS2		got;
	int		i;
	unsigned char uh;
	unsigned char ul;
	UCS2		value;

	initialize (); 
	init (in, _size);

	if (map==0)
	{
		for (i=0; i<inputSize; i++)
		{
			value = in[i];
			if (value>0xff)
			{
				append16 (value);
				continue;
			}
			append ((unsigned char) value);
		}
		*retSize = size;
		terminateStr ();
		return uchar;
	}

	NStateModel model(1, map->getEncodeInWordSize (),
		map->getEncodeOutWordSize (), 1);

	NBuffer	buffer(1,1);

	int stt, j;

	for (i=0; i<inputSize; i++)
	{
		stt = map->encode (&in[i], 1, &model, (i<inputSize-1) ? 1 : 0);

		buffer.append (&in[i], 1);

		if (stt == NM_ACCEPT) continue;
		if (stt >=0)
		{
			for (j=0; j<model.out.length; j++)
			{
				value=(UCS2) model.out.u.u16[j];
				if (map->getEncodeOutWordSize ()>0)
				{
					append ((const unsigned char)((value>>8)&0xff));
					append ((const unsigned char)(value&0xff));
				}
				else
				{
					append ((const unsigned char)value);
				}
				
			}
			i = i - (buffer.length-stt);
			buffer.reset();
			model.reset();
			continue;
		}
		// NM_REJECT
		if (map->getEncodeOutWordSize ()>0)
		{
			uh = (buffer.u.u16[0] >> 8) & 0xff;
			ul = buffer.u.u16[0] & 0xff;
			append (uh);
			append (ul);
		}
		else
		{
			// We should quote?
			if (buffer.u.u16[0] > 0xff)
			{
				append16 ((UCS2)buffer.u.u16[0]);
			}
			else
			{
				append ((const unsigned char) buffer.u.u16[0]);
			}
		}
		i = i - (buffer.length - 1);
		buffer.reset();
		model.reset();
	}
	*retSize = size;
	terminateStr ();
	return uchar;
}

U16BitConv::U16BitConv (const char* mapName, Endian en) 
	: UCS2Converter (mapName)
{
	endian = en;
}

U16BitConv::~U16BitConv()
{
}

unsigned char*
U16BitConv::encode (const UCS2 *in, int _size, int* retSize)
{
	int		i;
	unsigned char	ucs[2];
	init 		(in, _size);

	for (i=0; i<inputSize; i++)
	{
		// 12 34 -> 12 34
		if (endian==UCS2BE)
		{
			ucs[0] = in[i]>>8;
			ucs[1] = in[i]&0xff;
		}
		else // 12 34 -> 34 12
		{
			ucs[1] = in[i]>>8;
			ucs[0] = in[i]&0xff;
		}

		append (ucs[0]);
		append (ucs[1]);
	}
	*retSize = size;
	terminateStr();
	return uchar;
}

UCS2*
U16BitConv::decode (const unsigned char *in, int _size, int* retSize)
{
	int	i;
	UCS2	ucs;

	init (in, _size);
	for (i=0; i<inputSize-1; i=i+2)
	{
		if (endian==UCS2BE)
		{
			ucs = (in[i] << 8) | in[i+1];
		}
		else // 12 34 -> 34 12
		{
			ucs = (in[i+1] << 8) | in[i];
		}
		append (ucs);
	}
	if (inputSize % 2 != 0) append ((UCS2) '?');
	terminateUCS2 ();
	*retSize = size-1;
	return ucs2;
}

U16BEConv::U16BEConv(const char* mapName) : U16BitConv (mapName, UCS2BE)
{
}

U16BEConv::~U16BEConv ()
{
}

U16LEConv::U16LEConv(const char* mapName) : U16BitConv (mapName, UCS2LE)
{
}

U16LEConv::~U16LEConv ()
{
}

//
//
//
Uuniconv::Uuniconv (const char* mapName) : UCS2Converter (mapName)
{
}

Uuniconv::~Uuniconv()
{
}

//
// This wiil always produce big endian output.
// It puts \r at the end of line.
//
unsigned char*
Uuniconv::encode (const UCS2 *in, int _size, int* retSize)
{
	int		i;
	unsigned char	ucs[2];
	init 		(in, _size);

	endian = UCS2BE;
	if (endian==UCS2BE)
	{
		ucs[0] = UNI_MAGIC >>8;
		ucs[1] = UNI_MAGIC & 0xff;
	}
	else // 12 34 -> 34 12
	{
		ucs[1] = UNI_MAGIC>>8;
		ucs[0] = UNI_MAGIC&0xff;
	}

	append (ucs[0]);
	append (ucs[1]);
	for (i=0; i<inputSize; i++)
	{
		// 12 34 -> 12 34
		if (endian==UCS2BE)
		{
			ucs[0] = in[i]>>8;
			ucs[1] = in[i]&0xff;
		}
		else // 12 34 -> 34 12
		{
			ucs[1] = in[i]>>8;
			ucs[0] = in[i]&0xff;
		}
		append (ucs[0]);
		append (ucs[1]);
	}
	*retSize = size;
	terminateStr();
	return uchar;
}

//
// Endiannness is decided from beginning of file.
// line  separator paragraph seperator and \r is 
// translated.
UCS2*
Uuniconv::decode (const unsigned char *in, int _size, int* retSize)
{
	int	i;
	UCS2	ucs;

	init (in, _size);

	// set some defualt state.
	endian = UCS2BE;
	for (i=0; i<inputSize-1; i=i+2)
	{
		if (in[i] == (UNI_MAGIC>>8) 
			&& in[i+1] == (UNI_MAGIC&0xff))
		{
			endian=UCS2BE;
			continue;
		}
		if (in[i+1] == (UNI_MAGIC>>8) 
			&& in[i] == (UNI_MAGIC&0xff))
		{
			endian=UCS2LE;
			continue;
		}
		if (endian==UCS2BE)
		{
			ucs = (in[i] << 8) | in[i+1];
		}
		else // 12 34 -> 34 12
		{
			ucs = (in[i+1] << 8) | in[i];
		}
		append (ucs);
		
	}
	if (inputSize % 2 != 0) append ((UCS2) '?');
	terminateUCS2 ();
	*retSize = size-1;
	return ucs2;
}

UMSTXTConv::UMSTXTConv (const char* mapName) : UCS2Converter (mapName)
{
}

UMSTXTConv::~UMSTXTConv()
{
}

//
// This wiil always produce little endian output.
// It puts \r at the end of line.
//
unsigned char*
UMSTXTConv::encode (const UCS2 *in, int _size, int* retSize)
{
	int		i;
	unsigned char	ucs[2];
	unsigned char	ucsr[2];
	init 		(in, _size);

	endian = UCS2LE;
	if (endian==UCS2BE)
	{
		ucs[0] = UNI_MAGIC >>8;
		ucs[1] = UNI_MAGIC & 0xff;
	}
	else // 12 34 -> 34 12
	{
		ucs[1] = UNI_MAGIC>>8;
		ucs[0] = UNI_MAGIC&0xff;
	}

	append (ucs[0]);
	append (ucs[1]);
	ucsr[0] = (endian==UCS2BE) ?  0 : '\r';
	ucsr[1] = (endian==UCS2BE) ?  '\r' : 0;
	for (i=0; i<inputSize; i++)
	{
		// 12 34 -> 12 34
		if (endian==UCS2BE)
		{
			ucs[0] = in[i]>>8;
			ucs[1] = in[i]&0xff;
		}
		else // 12 34 -> 34 12
		{
			ucs[1] = in[i]>>8;
			ucs[0] = in[i]&0xff;
		}

		if (in[i]=='\n')
		{
			append (ucsr[0]);
			append (ucsr[1]);
		}
		append (ucs[0]);
		append (ucs[1]);
	}

	// HUh. Hope nobody wants to cut&paste 
	if (i==0 || in[i-1]!='\n')
	{
		append (ucsr[0]);
		append (ucsr[1]);
	}

	*retSize = size;
	terminateStr();
	return uchar;
}

//
// Endiannness is decided from beginning of file.
// line  separator paragraph seperator and \r is 
// translated.
UCS2*
UMSTXTConv::decode (const unsigned char *in, int _size, int* retSize)
{
	int	i;
	UCS2	ucs;

	init (in, _size);

	// set some defualt state.
	endian = UCS2LE;
	for (i=0; i<inputSize-1; i=i+2)
	{
		if (in[i] == (UNI_MAGIC>>8) 
			&& in[i+1] == (UNI_MAGIC&0xff))
		{
			endian=UCS2BE;
			continue;
		}
		if (in[i+1] == (UNI_MAGIC>>8) 
			&& in[i] == (UNI_MAGIC&0xff))
		{
			endian=UCS2LE;
			continue;
		}
		if (endian==UCS2BE)
		{
			ucs = (in[i] << 8) | in[i+1];
		}
		else // 12 34 -> 34 12
		{
			ucs = (in[i+1] << 8) | in[i];
		}
		switch (ucs)
		{
		case '\r':
			break;
		case 0x2029: // paragraph separator
		case 0x2028: // line separator
			append ((UCS2)'\n');
			break;
		default:
			append (ucs);
		}
		
	}
	if (inputSize % 2 != 0) append ((UCS2) '?');
	terminateUCS2 ();
	*retSize = size-1;
	return ucs2;
}
