/** 
 *  Yudit Unicode Editor Source File
 *
 *  Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "swidget/SLayout.h"
#include <stdio.h>

SLayout::SLayout (void) : lc(0,0), rc(0,0), dlc(0,0), drc(0,0)
{
}

/**
 * The layout looks like this:
 * lc-------------+
 * |              |
 * |              |
 * +-------------rc
 * When resize happens the lc is moved by dlc *(lc-newlc)
 * so these are deltas. The unit for deltas are 1000, so resizing
 * by unit 1 is achieved by 100.
 * Example a component that has a minimum size of 100 and grows 
 * twice when resized:
 * lc.x lc.y  rc.y  rc.y  dlc.x dlc.y  drc.x drc.y
 * 0,   0,    10,   10     0,   0      100,  100
 * it will also move down if:
 * 0,   0,    10,   10     0,   100    100,  100
 * it will also move right if:
 * 0,   0,    10,   10     100,  0     100,  100
 */
SLayout::SLayout (const SLocation& _lc, const SLocation _rc,
       const  SLocation& _dlc, const SLocation _drc) : 
  lc (_lc), rc (_rc), dlc(_dlc), drc(_drc)
{
}

SLayout::SLayout (const SLocation& _lc, const SLocation _rc) 
   : lc (_lc), rc (_rc)
{
  dlc = SLocation (100, 100);
  drc = SLocation (100, 100);
}

/**
 * copy the layout.
 */
SLayout::SLayout (const SLayout& l)
{
  lc = l.lc;
  rc = l.rc;
  dlc = l.dlc;
  drc = l.drc;
}
SLayout
SLayout::operator = (const SLayout& l)
{
  lc = l.lc;
  rc = l.rc;
  dlc = l.dlc;
  drc = l.drc;
  return *this;
}

SLayout::~SLayout ()
{
}

/**
 * Add layout to this to make a bigger layout.
 */
SLayout
SLayout::operator += (const SLayout l)
{
  if (l.rc.x > rc.x) rc.x = l.rc.x;
  if (l.rc.y > rc.y) rc.y = l.rc.y;
  return *this;
}

bool
SLayout::isEmpty() const
{
  return (lc == rc);
}

/**
 * This is in fact the minimum size.
 */
SDimension
SLayout::getDimension () const
{
  SDimension r (rc.x, rc.y);
  SDimension l (lc.x, lc.y);
  return SDimension (r-l);
  //return SDimension (r);
}

/**
 * Suppose in is the parent layout. we are resized.
 * @return the new location of this child component.
 */ 
SLocation 
SLayout::getLocation (const SLayout& parent, const SDimension& d) const
{
  SLocation l(d.width, d.height);
  SLocation delta = l - (parent.rc-parent.lc);
//fprintf (stderr, "dim=%u,%u delta=%d,%d parent=%d,%d\n", 
 //    d.width, d.height,
  //    delta.x, delta.y, parent.rc.x, parent.rc.y);
  SLocation lnew (lc.x + (delta.x * dlc.x) / 100, 
      lc.y + (delta.y * dlc.y) / 100);
  if (lnew.x < lc.x) lnew.x = lc.x;
  if (lnew.y < lc.y) lnew.y = lc.y;
  return SLocation (lnew);
}

/**
 * Suppose in is the parent layout. we are resized.
 * @return the new dimension of this child component.
 */ 
SDimension 
SLayout::getDimension (const SLayout& parent, const SDimension& d) const
{
  SLocation l(d.width, d.height);
  SLocation delta = l - (parent.rc - parent.lc);
  SLocation lnew (rc.x + (delta.x * drc.x) / 100, 
     rc.y + (delta.y * drc.y) / 100);
  if (lnew.x <= rc.x) lnew.x = rc.x +1;
  if (lnew.y <= rc.y) lnew.y = rc.y +1;
  SLocation left = getLocation (parent, d);
  return SDimension (lnew.x - left.x, lnew.y - left.y);
}
bool
SLayout::operator == (const SLayout l) const
{
  return (lc == l.lc && rc == l.rc && dlc == l.dlc && drc == l.drc);
}

bool
SLayout::operator != (const SLayout l) const
{
  return (lc != l.lc || rc != l.rc || dlc != l.dlc || drc != l.drc);
}

#if 0
void
SLayout::debug () const
{
  //fprintf (stderr,"lc=%d,%d rc=%d,%d dlc=%d,%d drc=%d,%d.\n",
   //    lc.x, lc.y, rc.x, rc.y, dlc.x, dlc.y, drc.x, drc.y);
}
#endif
