/** 
 *  Yudit Unicode Editor Source File
 *
 *  Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include "swindow/SFont.h"

#include "swindow/SFontFB.h"
#include "stoolkit/SMatrix.h"
#include "stoolkit/SHashtable.h"
#include "stoolkit/STextData.h"

static SFontFB fallbackFont;
static SHashtable<SFontImplVector> fontHashtable;

static SFontFB::SIGN mapFB (const SGlyph& glyph);

/* write the composing character bellow this */

/* if you see 0x7FFFFFFF then it is not used. */
#define  SD_USUALLY_HAS_DECOMPOSE 0x7FFFFFFF

/**
 * Create a font. The font can have many faces. 
 * @param _name is the logical name of this font.
 */
SFont::SFont (const SString _name) : name(_name), xlfd (SD_XLFD_ANY)
{
 if (fontHashtable.get (name) != 0) fontVector = fontHashtable[name];
 setSize (16.0);
}

/**
 * get a point 16 font called _default
 */
SFont::SFont (void) : name("default"), xlfd (SD_XLFD_ANY)
{
  static bool _defaultSet=false;
  if (_defaultSet==false)
  {
    SStringVector m(
        //"timesbi.ttf,arial.ttf,cyberbit.ttf,"
        "-*-*-medium-r-normal--16-*-*-*-c-*-iso8859-1,"
        "-*-*-*-*-*--16-*-*-*-c-*-iso8859-1,"
        "arial.ttf,cyberbit.ttf,"
    );
    SFontImplVector list;
    for (unsigned int i=0; i<m.size(); i++)
    {
       SFontImpl impl (m[i], m[i]);
       list.append (impl);
    }
    put ("default", list);
  }
  fontVector = fontHashtable["default"];
  setSize (16.0);
  
}

/**
 * Copy this font.
 */
SFont::SFont (const SFont& font)
{
  name = font.name;
  xlfd = font.xlfd;
  fontAscent = font.fontAscent;
  fontDescent = font.fontDescent;
  fontWidth = font.fontWidth;
  fontGap = font.fontGap;
  fontVector = font.fontVector;
  fontScale = font.fontScale;
}

/**
 * assign this font.
 */
SFont
SFont::operator = (const SFont& font)
{
  if (&font == this) return *this;
  name = font.name;
  fontAscent = font.fontAscent;
  fontDescent = font.fontDescent;
  fontWidth = font.fontWidth;
  fontGap = font.fontGap;
  xlfd = font.xlfd;
  fontScale = font.fontScale;
  fontVector = font.fontVector;
  return *this;
}

SFont::~SFont()
{
}


/**
 * Go through the list and set the size.
 */
void
SFont::setSize(double points)
{
  fontScale = points;
  SS_Matrix2D m;
  double sc = fallbackFont.scale();
  m.scale (fontScale * sc, fontScale * sc);
  fontGap =  fallbackFont.gap(m);
  fontAscent =  fallbackFont.ascent(m);
  fontDescent =  fallbackFont.descent(m);

  for (unsigned int i=0; i<fontVector.size(); i++)
  {
    SFontImpl im = fontVector[i];
    im.scale (fontScale, fontScale);
    //fontVector.remove (i);
    fontVector.replace (i, im);
    double g = im.gap ();
    double a = im.ascent ();
    double d = im.descent ();
    if (g > fontGap) fontGap = g;
    if (a > fontAscent) fontAscent = a;
    if (d > fontDescent) fontDescent = d;
  }
}

/**
 * Get the size of the font.
 * @return the size in points
 */
double
SFont::getSize () const
{
  return fontScale;
}

/**
 * A static method to build a font list.
 * This is only use at initialization time.
 * @param name is the name of the font.
 * @param gface is the face to add to the name.
 */
void
SFont::put (const SString name, const SFontImplVector& face)
{
  fontHashtable.put (name, face);
}

/**
 * clear all the stuff in the list
 */
void
SFont::clear()
{
}

/**
 * map a fallback font sign.
 */
static SFontFB::SIGN
mapFB (const SGlyph& glyph)
{
  if (glyph.size() > 1)
  {
    if (glyph[0] == SGlyph::CR || glyph[1] == SGlyph::LF) return SFontFB::CRLF;
    if (glyph[1] == SGlyph::CR || glyph[0] == SGlyph::LF) return SFontFB::LFCR;
  }
  else
  {
    if (glyph[0] == SGlyph::CR) return SFontFB::CR;
    if (glyph[0] == SGlyph::LF) return SFontFB::LF;
    if (glyph[0] == SGlyph::LS) return SFontFB::LS;
    if (glyph[0] == SGlyph::PS) return SFontFB::PS;
    if (glyph[0] == SGlyph::TAB) return SFontFB::TAB;
    if (glyph[0] < 32) return SFontFB::CTRL;
  }
  return SFontFB::LF;
}

/**
 * Try to draw one single glyph.
 * @param canvas is the canvas to draw to 
 * @param m is the conversion matrix
 * @param uch is the array containing ucs4 
 * @prama len is the length of the array
 */
void
SFont::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m, 
  const SGlyph& glyph)
{
  if (glyph.size() == 0)
  {
    return;
  }
  if (glyph.isSpecial())
  {
    SS_Matrix2D sd;
    double sc = fallbackFont.scale();
    sd.scale (fontScale * sc, fontScale * sc);
    SS_Matrix2D sm =  m * sd;
    SFontFB::SIGN sig = mapFB (glyph);
    fallbackFont.signDraw(canvas, pen, sm, sig, glyph[0]);
    return;
  }
  /* try to draw it with all fonts. */

  /* first try the precomposed */
  if (glyph.composition != 0 && glyph.composition < SD_USUALLY_HAS_DECOMPOSE)
  {
    for (unsigned int i=0; i<fontVector.size(); i++)
    {
      SFontImpl im = fontVector[i];
      unsigned int  ret = 0;
      ret = im.draw (canvas, pen, m, &glyph.composition, 1);
//Hebrew.kmap a
#if 0
if (0xfb2e == glyph.composition)
{
  fprintf (stderr, "0xfb2e glyph.composition = %u\n", ret);
}
#endif
    if (ret > 0)
      if (ret > 0) 
      {
         return;
      }
    }
  }

  /* now try whatever we have */
  unsigned int index = 0;
  bool found1 = false;
  unsigned int i=0;
  while (i<fontVector.size())
  {
    SFontImpl im = fontVector[i];
    unsigned int  ret = 0;
    ret = im.draw (canvas, pen, m, &glyph.decomposition()[index], 
        glyph.size() - index);

//Hebrew.kmap a
#if 0
if (0xfb2e == glyph.composition)
{
  fprintf (stderr, "0xfb2e=%u [%u]\n", ret, glyph.decomposition()[index]);
}
#endif
    if (ret > 0)
    {
      index += ret;
      found1 = true;
      i = 0;
    }
    else
    {
      i++;
    }
    if (index >= glyph.size()) return;
  }
  if (found1) return;

  SS_Matrix2D sd;
  double sc = fallbackFont.scale();
  sd.scale (fontScale * sc, fontScale * sc);
  SS_Matrix2D sm =  m * sd;

  if (glyph.composition != 0)
  {
    fallbackFont.draw(canvas, pen, sm, glyph.composition);
  }
  else
  {
    for (unsigned int i=0; i< glyph.size(); i++)
    {
      fallbackFont.draw(canvas, pen, sm, glyph[i]);
    }
  }
  return;
}


/**
 * return the width of the characters
 * @param m is the conversion matrix
 * @param uch is the array containing ucs4 
 * @prama len is the length of the array
 */
 
double
SFont::width (const SGlyph& glyph)
{
  if (glyph.size() == 0)
  {
    return 0.0;
  }
  double maxw = 0.0;
  double currw = 0.0;
  if (glyph.isSpecial ())
  {
    SS_Matrix2D m;
    double sc = fallbackFont.scale();
    m.scale (fontScale * sc, fontScale * sc);
    SFontFB::SIGN sig = mapFB (glyph);
    currw =  fallbackFont.signWidth (m, sig);
    return currw;
  }
  /* first try the precomposed */
  if (glyph.composition != 0 && glyph.composition < SD_USUALLY_HAS_DECOMPOSE)
  {
    for (unsigned int i=0; i<fontVector.size(); i++)
    {
      SFontImpl im = fontVector[i];
      unsigned int u =0;
      currw = im.width (&glyph.composition, 1, &u);
      if (u>0 && currw >0.0)
      {
//fprintf (stderr, "glyph.composition %x\n", (unsigned int) glyph.composition);
        return currw;
      }
    }
  }
  /* now try whatever we have */
  unsigned int index = 0;
  bool found1 = false;
  unsigned int i=0;
  while (i<fontVector.size())
  {
    SFontImpl im = fontVector[i];
    unsigned int u =0;
    currw = im.width (&glyph.decomposition()[index], glyph.size() - index, &u);

    if (u>0 && currw >0.0)
    {
      if (currw > maxw) maxw = currw;
      index += u;
      i = 0;
      found1 = true;
    }
    else
    {
      i++;
    }
    if (index >= glyph.size())
    {
      if (maxw > 0.0) return maxw;
      break;
    }
  }
  if (found1) return maxw;

  SS_Matrix2D sm;
  double sc = fallbackFont.scale();
  sm.scale (fontScale * sc, fontScale * sc);
 
  if (glyph.composition != 0)
  {
    maxw = fallbackFont.width (sm, glyph.composition);
  }
  else
  {
    for (unsigned int i=0; i< glyph.size(); i++)
    {
      currw = fallbackFont.width (sm, glyph[i]);
      if (currw > maxw) maxw = currw;
    }
  }
  return maxw;
}

/**
 * return the overall width
 */
double
SFont::width () const
{
  return fontWidth;
}

/**
 * return the overall ascent
 */
double
SFont::ascent () const
{
  return fontAscent;
}


/**
 * return the overall descent
 */
double
SFont::descent () const
{
  return fontDescent;
}

/**
 * return the overall gap
 */
double
SFont::gap () const
{
  return fontGap;
}
