/** 
 *  Yudit Unicode Editor Source File
 *
 *  Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include "swindow/SFontImpl.h"
#include "swindow/SFontTTF.h"
#include "swindow/SFontNative.h"
#include "swindow/SAwt.h"
#include "swindow/SImage.h"

#include "stoolkit/SIO.h"
#include "stoolkit/SBinHashtable.h"
#include "stoolkit/SHashtable.h"
#include "stoolkit/SExcept.h"
#include "stoolkit/SUtil.h"

static SStringVector fontSearchPath(
   "/,fonts,../fonts,/etc/fonts,/usr/share/yudit/fonts");


/* it makes a big difference in speed */

/**
 * @author: Gaspar Sinai <gsinai@yudit.org>
 * @version: 2000-04-23
 * This is the abstract widget toolkit font package
 */


class SFontDelegate
{
public:
  enum SE_Type { SE_NONE, SE_TTF, SE_NATIVE };
  SFontDelegate (const SFile& file, SE_Type type);
  ~SFontDelegate();
  unsigned int count;
  SE_Type type;
  union {
   SFontTTF* ttf;
   SFontNative* native;
  } u;
};

/**
 * Create a brand new delegate
 */
SFontDelegate::SFontDelegate (const SFile& file, SE_Type _type)
{
  count = 1;
  type = _type;
  u.ttf = 0;
  switch (type)
  {
  case SE_TTF:
    u.ttf = new SFontTTF (file);
    if (!u.ttf->isOK())
    {
      delete u.ttf;
      u.ttf = 0;
      type = SE_NONE;
    }
    break;
  case SE_NATIVE:
    u.native = SAwt::getFont();
    break;
  case SE_NONE:
     break;
  }
}

SFontDelegate::~SFontDelegate()
{
  switch (type)
  {
  case SE_TTF:
     if (u.ttf!=0) delete u.ttf;
     break;
  case SE_NATIVE:
     if (u.native!=0) delete u.native;
     break;
  case SE_NONE:
     break;
  }
}

static unsigned int fontCacheUsers=0;
typedef SBinHashtable<SFontDelegate*> SFontCache;
static SFontCache* fontCache=0;

/**
 * Create a new font face.
 * @param _xlfd is the logical x11 font descriptor that matches
 *        this font.  "-%s-%s-%s-%s-*-%s-%g-*-*-*-%s-%s-%s-%s"
 * foundry-family-weight-slant-*-style-pixel-*-*-*-spacing-avgWidth-registry-encoding
 */
SFontImpl::SFontImpl (const SString& _name, const SString& _xlfd) : 
    name (_name), xlfd (_xlfd)
{
  /**
   * Try to find TTF first
   */
  if (fontCacheUsers==0)
  {
     fontCache = new SFontCache();
  }
  fontCacheUsers++;
  //fprintf (stderr, "XLFD %*.*s\n", SSARGS(xlfd));
  SFontDelegate* d = fontCache->get(name);
  delegate = 0;
  if (d!=0)
  {
    //fprintf (stderr, "resusing %*.*s\n", SSARGS(name));
    delegate = d;
    d->count++;
    if (d->type == SFontDelegate::SE_NATIVE)
    {
      createSaneXLFD();
    }
  }
  else
  {
    SString n (name);
    //n.append (".ttf");
    SFile f(n, fontSearchPath);
    if (f.size() > 0)
    {
      d = new SFontDelegate (f, SFontDelegate::SE_TTF);
      CHECK_NEW (d);
      delegate = d;
      // It is obvious that we dont use fontCache at this point.
      // FIXME: this was commented out till yudit-2.3.beta-14 !!! Why?
      fontCache->put (name, d);
    }
    else
    {
      createSaneXLFD();
      d = new SFontDelegate (xlfd, SFontDelegate::SE_NATIVE);
      CHECK_NEW (d);
      delegate = d;
      fontCache->put (name, d);
    }
  }
}
void
SFontImpl::createSaneXLFD ()
{
  SStringVector l;
  xlfd.replace("--", "-*-");
  l.split (xlfd, "-");
  if (l.size() < 14)
  {
    //fprintf (stderr, "SFontImpl: Need better XLFD %*.*s. Adding some -*-\n", SSARGS(xlfd));
    /* find a good insertion point - FIXME this is not very good here */
    unsigned int ipoint = 0;
    if (l.size() > 2)
    {
      ipoint = l.size() -1;
      while (ipoint > 0 && l[ipoint-1] != "*") ipoint--;
    }
    else if (xlfd[0] == '-')
    {
      while (ipoint < l.size()  && l[ipoint] != "*") ipoint++;
      if (ipoint >= l.size())
      {
        ipoint = 0;
      }
    }
    while (l.size() < 14) l.insert(ipoint, "*");
    xlfd = "-";
    xlfd.append(l.join ("-"));
  }
}

SFontImpl::SFontImpl (const SFontImpl& ff)
{
  if (fontCacheUsers==0)
  {
     fontCache = new SFontCache();
  }
  fontCacheUsers++;
  name = ff.name;
  xlfd = ff.xlfd;
  matrix = ff.matrix;
  SFontDelegate* d = (SFontDelegate*) ff.delegate;
  if (d)
  {
    d->count++;
  }
  delegate = d;
}

SFontImpl
SFontImpl::operator= (const SFontImpl& ff)
{
  if (&ff == this) return *this;
  name = ff.name;
  xlfd = ff.xlfd;
  matrix = ff.matrix;
  SFontDelegate* d;
  if (delegate)
  {
    d = (SFontDelegate*) delegate;
    d->count--;
    if (d->count==0) delete d;
  }
  d = (SFontDelegate*) ff.delegate;
  if (d)
  {
    d->count++;
  }
  delegate = d;
  return *this;
  
}
SFontImpl::~SFontImpl ()
{
  if (delegate)
  {
    SFontDelegate* d = (SFontDelegate*) delegate;
    d->count--;
    if (d->count==0)
    {
       fontCache->remove (name);
       delete d;
    }
  }
  fontCacheUsers--;
  if (fontCacheUsers==0)
  {
    delete fontCache;
    fontCache = 0;
  }
}

/**
 * set the search path for the font
 * This path will be used to locate font files on the disk
 * Currently it is used for True Type fonts.
 * @param l is the list of directories.
 */
void
SFontImpl::setPath(const SStringVector& l)
{
//fprintf (stderr, "set path ...\n");
  fontSearchPath = l;
}

/**
 * search files for property in order and set the path to the 
 * property. Always add YUDIT_DATA/fonts
 */
void
SFontImpl::guessPath (const SStringVector& files, const SString& property)
{
  
  SStringVector outDataPath;
  for (unsigned int i=0; i<files.size(); i++)
  {
    SProperties p;
    loadProperties (files[i], &p);
    if (p.get (property))
    {
      SStringVector v(p[property], ",:;");
      for (unsigned int j=0; j<v.size(); j++)
      {
         outDataPath.append (v[j]);
      }
    }
  }
  SString c1 = getHome();
  c1.append ("/.yudit/fonts");
  SString c2 = getPrefix();
  c2.append ("/fonts");
  if (outDataPath.size()!=0)
  { 
    outDataPath.append (c1); 
    outDataPath.append (c2); 
    fontSearchPath = outDataPath;
  }
  else
  {
    outDataPath.append (c1); 
    outDataPath.append (c2); 
    outDataPath.append (fontSearchPath);
    fontSearchPath = outDataPath;
  }
//fprintf (stderr, "fontpath is %*.*s\n", SSARGS(fontSearchPath.join(",")));
}
/**
 * get the path using the "yudit.default.path"
 */
void
SFontImpl::guessPath()
{
  SString c1 = getHome();
  c1.append ("/.yudit/yudit.properties");
  SString c2 = getPrefix();
  c2.append ("/config/yudit.properties");
  SStringVector v;
  v.append (c1);
  v.append (c2);
  guessPath (v, "yudit.fontpath");
}

/**
 * Scale the font to the given pointsize.
 * @param x is the x axis pointsize
 * @param y is the y axis pointsize
 */
void
SFontImpl::scale (double x, double y)
{
  SS_Matrix2D m;
  matrix = m;
  if (!delegate) return;
  SFontDelegate* d = (SFontDelegate*) delegate;;
  SStringVector l;
  char tmp[64];
  switch (d->type)
  {
  case SFontDelegate::SE_TTF:
    if (d->u.ttf->isOK())
    {
        double scale = d->u.ttf->scale ();
        m.scale (x * scale, y * scale);
        matrix = m;
    }
    break;
  case SFontDelegate::SE_NATIVE:
    /*
    * Modify the currentXLFD and load the font.
    * it will be our pixel size. nice eh? 
    * A reminder on xlfd:
    * -foundry-family-weight-slant-*-style-pixel-*-*-*-spacing-avgWidth-registry-encoding
    */
    if (d->u.native==0) break;
    l.split (xlfd, "-");
    sprintf (tmp, "%u", (unsigned int) (y + 0.5));
    if (l.size() < 14)
    {
      fprintf (stderr, "BAD XLFD %*.*s\n", SSARGS(xlfd));
      break;
    }
    l.replace (6, tmp);
    xlfd = "-";
    xlfd.append(l.join ("-"));
    break;
  case SFontDelegate::SE_NONE:
    break;
  }
  
}

/**
 * Try to draw one single glyph.
 * @param canvas is the canvas to draw to 
 * @param m is the conversion matrix
 * @param uch is the array containing ucs4 
 * @prama len is the length of the array
 * return the advance length
 * BUG: unfortunatelly this matrix is already normalized to one.....
 * scaling will return the same image. The reason is that the canvas is scaled
 * and not the matrix. Matrix should be scaled.
 */
unsigned int
SFontImpl::draw (SCanvas* canvas, const SPen& pen, const SS_Matrix2D& m, 
  const SS_UCS4* uch, unsigned int len)
{
  if (!delegate) return 0;

  SFontDelegate* d = (SFontDelegate*) delegate;;
  unsigned int used;
  SS_Matrix2D sd = m * matrix;
  switch (d->type)
  {
  case SFontDelegate::SE_TTF:
    if (d->u.ttf->isOK())
    {
        used = d->u.ttf->length (uch, len);
        if (used == 0) return 0;
        long larr[8];
        memset (larr, 0, sizeof (long) * 8);
        larr[0] = 'T' ; larr[1] = 'T'; larr[2] = 'F';
        larr[3] = (long)d->u.ttf; larr[4] = (long) used;
        for (unsigned int i=0; i<used; i++)
        {
           larr[i+5] = (long)uch[i];
        }
        SString key((char*) larr, 8 * sizeof (long));

        /* FIXME: you should in fact use an instance counter. */
        //fprintf (stderr, "scale=%g,%g;", sd.x0, sd.y1);
        key.append ((long)(sd.x0*16000.0));
        key.append ((long)(sd.y1*16000.0));

//fprintf (stderr, "m.t0=%g m.t1=%g\n", m.t0, m.t1);
        if (!canvas->newpath (m.t0, m.t1, key))
        {
          d->u.ttf->draw (canvas, sd, uch, len);
        }
        canvas->fill (pen);
        return used;
    }
    break;
  case SFontDelegate::SE_NATIVE:
    if (d->u.native==0) break;
//fprintf (stderr, "native=*.*s\n", SSARGS(xlfd));
    return d->u.native->draw (xlfd, canvas, pen, sd, uch, len);
  case SFontDelegate::SE_NONE:
     break;
  }
  return 0;
}

/**
 * return the width of the characters
 * @param m is the conversion matrix
 * @param uch is the array containing ucs4 
 * @prama len is the length of the array
 */
 
double
SFontImpl::width (const SS_UCS4* uch,
      unsigned int len, unsigned int* used)
{
  if (!delegate) return 0.0;

  SFontDelegate* d = (SFontDelegate*) delegate;;
  double ret;
  switch (d->type)
  {
  case SFontDelegate::SE_TTF:
    if (d->u.ttf->isOK())
    {
        ret = d->u.ttf->width (matrix, uch, len, used);
        return ret;
    }
    break;
  case SFontDelegate::SE_NATIVE:
    if (d->u.native==0) break;
    return d->u.native->width (xlfd, uch, len, used);
  case SFontDelegate::SE_NONE:
     break;
  }
  return 0.0;
}

/**
 * return the overall width
 */
double
SFontImpl::width () const
{
  if (!delegate) return 0.0;

  SFontDelegate* d = (SFontDelegate*) delegate;;
  switch (d->type)
  {
  case SFontDelegate::SE_TTF:
    if (d->u.ttf->isOK())
    {
        return (d->u.ttf->width (matrix));
    }
    break;
  case SFontDelegate::SE_NATIVE:
    if (d->u.native==0) break;
    return d->u.native->width (xlfd);
  case SFontDelegate::SE_NONE:
     break;
  }
  return 0.0;
}

/**
 * return the overall ascent
 */
double
SFontImpl::ascent () const
{
  if (!delegate) return 0.0;

  SFontDelegate* d = (SFontDelegate*) delegate;;
  switch (d->type)
  {
  case SFontDelegate::SE_TTF:
    if (d->u.ttf->isOK())
    {
        return (d->u.ttf->ascent (matrix));
    }
    break;
  case SFontDelegate::SE_NATIVE:
    if (d->u.native==0) break;
    return d->u.native->ascent (xlfd);
  case SFontDelegate::SE_NONE:
     break;
  }
  return 0.0;
}


/**
 * return the overall descent
 */
double
SFontImpl::descent () const
{
  if (!delegate) return 0.0;

  SFontDelegate* d = (SFontDelegate*) delegate;;
  switch (d->type)
  {
  case SFontDelegate::SE_TTF:
    if (d->u.ttf->isOK())
    {
        return (d->u.ttf->descent (matrix));
    }
    break;
  case SFontDelegate::SE_NATIVE:
    if (d->u.native==0) break;
    return d->u.native->descent (xlfd);
  case SFontDelegate::SE_NONE:
     break;
  }
  return 0.0;
}

/**
 * return the overall gap
 */
double
SFontImpl::gap () const
{
  if (!delegate) return 0.0;

  SFontDelegate* d = (SFontDelegate*) delegate;;
  switch (d->type)
  {
  case SFontDelegate::SE_TTF:
    if (d->u.ttf->isOK())
    {
        return (d->u.ttf->gap (matrix));
    }
    break;
  case SFontDelegate::SE_NATIVE:
    if (d->u.native==0) break;
    return d->u.native->gap (xlfd);
  case SFontDelegate::SE_NONE:
     break;
  }
  return 0.0;
}

SObject*
SFontImpl::clone () const
{
  SFontImpl *ret =  new SFontImpl(*this);
  CHECK_NEW (ret);
  return ret;
}
