/** 
 *  Yudit Unicode Editor Source File
 *
 *  Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include "swindow/sx11/SX11Font.h"
#include "stoolkit/SBinHashtable.h"
#include "stoolkit/SExcept.h"
#include <X11/Xatom.h>

#include <stdio.h>

static SStringVector mapNameVector;
static bool inited= false;
static char* defaultNames[] =
{
"*-iso10646-1", "ucs2",
"*-unicode", "ucs2",
"*-iso8859-1", "iso-8859-1",
"*-iso8859-2", "iso-8859-2",
"*-iso8859-3", "iso-8859-3",
"*-iso8859-4", "iso-8859-4",
"*-iso8859-5", "iso-8859-5",
"*-iso8859-6", "iso-8859-6",
"*-iso8859-7", "iso-8859-7",
"*-iso8859-8", "iso-8859-8",
"*-iso8859-9", "iso-8859-9",
"*-koi8-r", "koi8-r",
"*-koi8-1", "koi8-r",

"*-ksc5601.1987-0", "ksc-5601-l",
"*-ksc5601*", "ksc-5601-l",

"*-jisx0208.1983-0", "jis-0208",
"*-jisx0208*", "jis-0208",

"*-jisx0208.1990-0", "jis-0208",
"*-jisx0208*", "jis-0208",

"*-jisx0201.1976-0", "jis-0201",
"*-jisx0201*", "jis-0201",

"*-jisx0212.1990-0", "jis-0212",
"*-jisx0212*-0", "jis-0212",

"*-big5.hku-0", "big-5",
"*-big5-0", "big-5",            
"*-big5*", "big-5",            

"*-gb2312.1980-0", "gb-2312-l", /* There are some font encoded with gb-2312-r"*/
"*-gb2312*", "gb-2312-l", /* There are some font encoded with gb-2312-r"*/

"*-iso646*", "iso-646", /* 646 is not 10646 */
};

/**
 * @author: Gaspar Sinai <gsinai@yudit.org>
 * @version: 2000-06-03
 * This is a specific native font, with a certain size
 */
class SX11FontCache
{
public:
  SX11FontCache (SX11Impl *impl,const SString& xlfd);
  ~SX11FontCache ();
  SString xlfd;
  SX11Impl* impl;
  SUniMap map;
  double fontAscent;
  double fontDescent;
  double fontWidth;
  XFontStruct* font;
  bool loaded;
  bool load ();
};


SX11FontCache::SX11FontCache(SX11Impl *_impl,  const SString& _xlfd) 
     : xlfd (_xlfd)
{
  impl = _impl;
  fontAscent = 0.0;
  fontDescent = 0.0;
  fontWidth = 0.0;
  loaded = false;
}

/**
 * Create an XLFD and try to load this as a font.
 * TODO: it should modify xlfd to show the real xlfd
 */
bool
SX11FontCache::load ()
{
  if (loaded) return (font!=0);
  loaded = true;
  char* fontName = xlfd.cString();
  font = XLoadQueryFont (impl->display, fontName);
  delete fontName;
  if (font == 0)  return false;
  /* Fix broken fonts */
  Atom fontatom = XInternAtom (impl->display, "FONT", False);
  char* vle = 0;
  XFontProp *fp = font->properties;
  for (int i = 0; i < font->n_properties; i++, fp++)
  {
    if (fp->name == fontatom)
    {
      vle = XGetAtomName (impl->display, fp->card32);
      break;
    }
  }
  if (vle)
  {
    SString fn(vle);
    fn.lower();
    xlfd = fn;
  }

  int tryAscent[2];
  tryAscent[0] = font->ascent;
  tryAscent[1] = font->max_bounds.ascent;
  if (tryAscent[1] < tryAscent[0]
     || (tryAscent[0] +2) * 2 < tryAscent[1])
  {
    tryAscent[1] = tryAscent[0];
  }
  int tryDescent[2];
  tryDescent[0] = font->descent;
  tryDescent[1] = font->max_bounds.descent;
  if (tryDescent[1] < tryDescent[0]
     || (tryDescent[0] +2) * 2 < tryDescent[1])
  {
    tryDescent[1] = tryDescent[0];
  }
  fontAscent = (double) tryAscent[1];
  fontDescent = (double)tryDescent[1];
  fontWidth = (double) font->max_bounds.width;
  if (!inited)
  {
    for (unsigned int i=0; i<sizeof (defaultNames) / sizeof (char*)/2; i++)
    {
      mapNameVector.append (defaultNames[2*i]);
      mapNameVector.append (defaultNames[2*i+1]);
    }
    inited = true;
  }
  /* Find out what map is used for this font */
  SString m;
  for (unsigned int i=0; i+1<mapNameVector.size(); i=i+2)
  {
    if (xlfd.match (mapNameVector[i]))
    {
      m = mapNameVector[i+1];
      break;
    }
  }
  if (m.size() ==0)
  {
    fprintf (stderr, "swindow/sx11/SX11Font need fontmap for %*.*s\n",
        SSARGS(xlfd));
  }
  if (m.size() !=0 && m != SString("ucs2"))
  {
    SUniMap um(m);
    if (!um.isOK())
    {
     fprintf (stderr, "SX11Font .my map '%*.*s' needed for for %*.*s\n",
        SSARGS(m), SSARGS(xlfd));
    }
    else
    {
      map = um;
    }
  }
  return true;
}

SX11FontCache::~SX11FontCache ()
{
  if (font) XFreeFont (impl->display, (XFontStruct*) font);
} 

/**
 * TODO: this is never freed up. It simply grows...
 */
static SBinHashtable<SX11FontCache*> fontCache;

static SX11FontCache* loadFont (SX11Impl* impl, const SString& xlfd);

SX11Font::SX11Font (SX11Impl* _impl) : impl(_impl)
{
}

SX11Font::~SX11Font ()
{
} 

unsigned int
SX11Font::draw (const SString& xlfd, SCanvas* canvas, const SPen& pen, 
    const SS_Matrix2D& matrix, const SS_UCS4* g, unsigned int len)
{
  SX11FontCache* fc = loadFont(impl, xlfd);
  if (!fc->load()) return 0;
  XFontStruct* font = fc->font;
  if (len==0) return 0;
  /* try to decode it */
  SS_UCS2 glyphno = fc->map.encode (g[0]);
  if (glyphno==0) return 0;

  XChar2b     char2B;
  char2B.byte1 = glyphno >> 8;
  char2B.byte2 = glyphno & 0xff;
  if (font->per_char==0)
  {
    int ddirection;
    int dascent;
    int ddescent;
    XCharStruct overall;
    if (!XQueryTextExtents16 (impl->display,
       font->fid, &char2B, 1, &ddirection, &dascent, &ddescent, &overall))
    {
      return 0;
    }
    if (overall.width==0) return 0;
  }
  else
  {
    if (char2B.byte1 < font->min_byte1 
      || char2B.byte2 < font->min_char_or_byte2
      || char2B.byte1 > font->max_byte1 
      || char2B.byte2 > font->max_char_or_byte2)
    {
       return 0;
    }
    int cellIndex = (char2B.byte1-font->min_byte1)
     * (font->max_char_or_byte2 -font->min_char_or_byte2+1)
     + char2B.byte2-font->min_char_or_byte2;
    if (font->per_char[cellIndex].width == 0) return 0;
  }
  canvas->bitfont (pen, matrix.t0, matrix.t1, 
         (void*) font->fid, (char*) &char2B, 2);
  return 1;
}

double
SX11Font::width (const SString& xlfd, const SS_UCS4* g, unsigned int len, unsigned int *sz)
{
  if (sz) *sz=0;
  SX11FontCache* fc = loadFont(impl, xlfd);
  if (!fc->load())return 0.0;
  XFontStruct* font = fc->font;
  if (len==0) return 0.0;
  /* try to decode it */
  SS_UCS2 glyphno = fc->map.encode (g[0]);
  if (glyphno==0) return 0.0;

  XChar2b     char2B;
  char2B.byte1 = glyphno >> 8;
  char2B.byte2 = glyphno & 0xff;
  if (font->per_char==0)
  {
    int ddirection;
    int ddescent;
    int dascent;
    XCharStruct overall;
    if (!XQueryTextExtents16 (impl->display,
       font->fid, &char2B, 1, &ddirection, &dascent, &ddescent, &overall))
    {
      return 0.0;
    }
    if (sz) *sz = 1;
    return (double) overall.width;
  }
  if (char2B.byte1 < font->min_byte1 || char2B.byte2 < font->min_char_or_byte2
      || char2B.byte1 > font->max_byte1 || char2B.byte2 > font->max_char_or_byte2)
  {
     return 0;
  }
  int cellIndex = (char2B.byte1-font->min_byte1)
     * (font->max_char_or_byte2 -font->min_char_or_byte2+1)
     + char2B.byte2-font->min_char_or_byte2;
  if (sz) *sz = 1;
  return (double) font->per_char[cellIndex].width;
  
}

double
SX11Font::width (const SString& xlfd)
{
  SX11FontCache* fc = loadFont(impl, xlfd);
  if (!fc->load()) return 0.0;
  return fc->fontWidth;
}

double
SX11Font::ascent (const SString& xlfd)
{
  SX11FontCache* fc = loadFont(impl, xlfd);
  if (!fc->load()) return 0.0;
  return fc->fontAscent;
}

double
SX11Font::descent (const SString& xlfd)
{
  SX11FontCache* fc = loadFont(impl, xlfd);
  if (!fc->load()) return 0.0;
  return fc->fontDescent;
}

double
SX11Font::gap (const SString& xlfd)
{
  return 0.0;
}

static SX11FontCache*
loadFont (SX11Impl* impl, const SString& xlfd)
{
  SX11FontCache* f = fontCache.get (xlfd);
#if 0
 if (xlfd.find("lucidabright") > 0) 
 {
  fprintf (stderr, "LUCIDA=%*.*s\n", SSARGS(xlfd));
 }
#endif
  if (f!=0) return f;
  f = new SX11FontCache (impl, xlfd);
  CHECK_NEW (f);
  fontCache.put (xlfd, f);
  return f;
}
